import { Hero6Utility } from "./hero6-utility.js";

export class Hero6RollDialog extends Dialog {

  /* -------------------------------------------- */
  static async create(actor, rollData) {

    let options = { classes: ["Hero6Dialog"], width: 320, height: 'fit-content', 'z-index': 99999 };
    let html = await renderTemplate('systems/fvtt-hero-system-6/templates/apps/roll-dialog-generic.hbs', rollData);

    return new Hero6RollDialog(actor, rollData, html, options);
  }

  /* -------------------------------------------- */
  constructor(actor, rollData, html, options, close = undefined) {
    let conf = {
      title: "Roll window",
      content: html,
      buttons: {
        roll: {
          icon: '<i class="fas fa-check"></i>',
          label: "Roll !",
          callback: () => { this.roll() }
        },
        cancel: {
          icon: '<i class="fas fa-times"></i>',
          label: "Cancel",
          callback: () => { this.close() }
        }
      },
      close: close
    }

    super(conf, options);

    this.actor = actor;
    this.rollData = rollData;
  }

  /* -------------------------------------------- */
  roll() {
    Hero6Utility.rollHero6(this.rollData)
  }

  /* -------------------------------------------- */
  async refreshDialog() {
    const content = await renderTemplate("systems/fvtt-hero-system-6/templates/apps/roll-dialog-generic.hbs", this.rollData)
    this.data.content = content
    this.render(true)
  }

  /* -------------------------------------------- */
  activateListeners(html) {
    super.activateListeners(html);

    let dialog = this;
    function onLoad() {
    }
    $(function () { onLoad(); });

    html.find('#advantage').change((event) => {
      this.rollData.advantage = event.currentTarget.value
    })
    html.find('#disadvantage').change((event) => {
      this.rollData.disadvantage = event.currentTarget.value
    })
    html.find('#rollAdvantage').change((event) => {
      this.rollData.rollAdvantage = event.currentTarget.value
    })
    html.find('#useshield').change((event) => {
      this.rollData.useshield = event.currentTarget.checked
    })
    html.find('#hasCover').change((event) => {
      this.rollData.hasCover = event.currentTarget.value
    })
    html.find('#situational').change((event) => {
      this.rollData.situational = event.currentTarget.value
    })
    html.find('#bonusMalus').change((event) => {
      this.rollData.bonusMalus = Number(event.currentTarget.value)
    })
    
  }
}